""" ClinicalOmicsDB: Bridging the gap between next-generation clinical omics data and machine learning
Reference: Chang In Moon, Byron Jia, Bing Zhang
Paper link: TBD
Code author: Chang In Moon (moonchangin@gmail.com)
-----------------------------

file_download.py
- API/SDK that will download data from box
"""
import threading
from queue import Queue

from werkzeug import Request
from werkzeug import Response
from werkzeug.serving import make_server

import webbrowser
import time
import logging
import os
import requests
from requests.adapters import HTTPAdapter, Retry
import getpass
import bs4

from file_tools import *
from exceptions import InvalidParameterError, NoInternetError, DownloadFailedError

# Some websites don't like requests from sources without a user agent. Let's preempt that issue.
USER_AGENT = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.10; rv:39.0)'
HEADERS = {'User-Agent': USER_AGENT}

def download_text(url):
    """Download text from a direct download url for a text file.

    Parameters:
    url (str): The direct download url for the text.

    Returns:
    str: The downloaded text.
    """
    
    try:
        s = requests.Session()
        retries = Retry(total=10,
                backoff_factor=0.1,
                status_forcelist=[ 500, 502, 503, 504 ])
        s.mount('http://', HTTPAdapter(max_retries=retries))
        response = s.get(url) #s.get(url, headers=HEADERS, allow_redirects=True)
        response.raise_for_status() # Raises a requests HTTPError if the response code was unsuccessful
    except requests.RequestException: # Parent class for all exceptions in the requests module
        raise NoInternetError("Insufficient internet. Check your internet connection.") from None 

    text = response.text.strip()
    return text


""" ClinicalOmicsDB: Bridging the gap between next-generation clinical omics data and machine learning
Reference: Chang In Moon. Byron Jia, Bing Zhang
Paper link: TBD
Code author: Chang In Moon (moonchangin@gmail.com)
-----------------------------

utils.py
- Various utility functions for ClinicalOmicsDB framework
(1) perf_metric: prediction performances in terms of AUROC or accuracy
"""

# Necessary packages
import numpy as np
from sklearn.metrics import accuracy_score, roc_auc_score

#%% 
def perf_metric (metric, y_test, y_test_hat):
  """Evaluate performance.
  
  Args:
    - metric: acc or auc
    - y_test: ground truth label
    - y_test_hat: predicted values
    
  Returns:
    - performance: Accuracy or AUROC performance
  """
  # Accuracy metric
  if metric == 'acc':
    if np.isnan(y_test_hat[:,1]).any() == True: # NaN predictions will be automatically be stored as 0
      result = 0
      return result
    result = accuracy_score(np.argmax(y_test, axis = 1), 
                            np.argmax(y_test_hat, axis = 1))
  # AUROC metric
  elif metric == 'auc':
    if np.isnan(y_test_hat[:,1]).any() == True: # NaN predictions will be automatically be stored as 0
      result = 0
      return result
    result = roc_auc_score(y_test, y_test_hat[:,1])
  return result



""" ClinicalOmicsDB: Bridging the gap between next-generation clinical omics data and machine learning
Reference: Chang In Moon, Byron Jia, Bing Zhang
Paper link: TBD
Code author: Chang In Moon (moonchangin@gmail.com)
-----------------------------
database_download.py
- Basic functions that will use box api to obtain dataset

(0) download_sheet: provide a table with list of dataset link
(1) get_dataset_link: provide available data in a data frame
(2) get_ssl_dataset_link: provide available data with multiple treatment arms
(3) HR_dataset_link: provide available data with BRCA-HR subtype
(4) HER2_dataset_link: provide available data with BRCA-HER2 subtype
(5) TNBC_dataset_link: provide available data with BRCA-TNBC subtype
(6) ssl_HR_dataset_link: provide available data with multiple treatment arms [BRCA-HR subtype]
(7) ssl_HER2_dataset_link: provide available data with multiple treatment arms [BRCA-HER2 subtype]
(8) ssl_TNBC_dataset_link: provide available data with multiple treatment arms [BRCA-TNBC subtype]
(9) download: returns dataset for analysis
(10) get_patient_dataset_link: provide available clinical in a data frame [under development]
(11) download_patient_data: returns dataset for subsetting data frame [under development]

"""
import io
import pandas as pd
from file_download import download_text as _download_text
from exceptions import BaseError, BaseWarning, InvalidParameterError, NoInternetError, OldPackageVersionWarning

def download_sheet():
    dataset_link_url = "https://bcm.box.com/shared/static/0t7n2lcx8pewasxjuicez78k9tyj0kc5.csv"
    try:
        dataset_link_text = _download_text(dataset_link_url)
    except NoInternetError:
        raise NoInternetError("Insufficient internet to download available dataset info. Check your internet connection.") from None
    return pd.read_csv(io.StringIO(dataset_link_text), header=0)


def get_dataset_link():
    # dataset_link_url = "https://bcm.box.com/shared/static/1rs6wid9em7tewjpqchnar4wlr75l3ml.csv"
    dataset_link_url = "https://bcm.box.com/shared/static/0t7n2lcx8pewasxjuicez78k9tyj0kc5.csv"
    try:
        dataset_link_text = _download_text(dataset_link_url)
    except NoInternetError:
        raise NoInternetError("Insufficient internet to download available dataset info. Check your internet connection.") from None
    return pd.read_csv(io.StringIO(dataset_link_text), header=0)

def get_ssl_dataset_link():
    dataset_link_url = "https://bcm.box.com/shared/static/6h0jdwq1lsfm13qmrfbv8yac8hje8zzb.csv"
    try:
        dataset_link_text = _download_text(dataset_link_url)
    except NoInternetError:
        raise NoInternetError("Insufficient internet to download available dataset info. Check your internet connection.") from None
    return pd.read_csv(io.StringIO(dataset_link_text), header=0)

def HR_dataset_link():
    dataset_link_url = "https://bcm.box.com/shared/static/57eq2myuuxpxjimm8tus5sjlhxijdgi1.csv"
    try:
        dataset_link_text = _download_text(dataset_link_url)
    except NoInternetError:
        raise NoInternetError("Insufficient internet to download available dataset info. Check your internet connection.") from None
    return pd.read_csv(io.StringIO(dataset_link_text), header=0)

def HER2_dataset_link():
    dataset_link_url = "https://bcm.box.com/shared/static/ifaxp6bxka909lggeb8fmyv4ze70ek6k.csv"
    try:
        dataset_link_text = _download_text(dataset_link_url)
    except NoInternetError:
        raise NoInternetError("Insufficient internet to download available dataset info. Check your internet connection.") from None
    return pd.read_csv(io.StringIO(dataset_link_text), header=0)

def TNBC_dataset_link():
    dataset_link_url = "https://bcm.box.com/shared/static/9hd1vmvcztsbxdzcw2ino2t4tmltw2hv.csv"
    try:
        dataset_link_text = _download_text(dataset_link_url)
    except NoInternetError:
        raise NoInternetError("Insufficient internet to download available dataset info. Check your internet connection.") from None
    return pd.read_csv(io.StringIO(dataset_link_text), header=0)

def ssl_HR_dataset_link():
    dataset_link_url = "https://bcm.box.com/shared/static/di0xm2y2w2n7f8hp7he4xgdugnadnf32.csv"
    try:
        dataset_link_text = _download_text(dataset_link_url)
    except NoInternetError:
        raise NoInternetError("Insufficient internet to download available dataset info. Check your internet connection.") from None
    return pd.read_csv(io.StringIO(dataset_link_text), header=0)

def ssl_HER2_dataset_link():
    dataset_link_url = "https://bcm.box.com/shared/static/hb2hgxogx2h0wq6ud0mbigjvrli5j4lx.csv"
    try:
        dataset_link_text = _download_text(dataset_link_url)
    except NoInternetError:
        raise NoInternetError("Insufficient internet to download available dataset info. Check your internet connection.") from None
    return pd.read_csv(io.StringIO(dataset_link_text), header=0)

def ssl_TNBC_dataset_link():
    dataset_link_url = "https://bcm.box.com/shared/static/8lfgpx9mkrc13fa0e9gfpsfjsoe29n2a.csv"
    try:
        dataset_link_text = _download_text(dataset_link_url)
    except NoInternetError:
        raise NoInternetError("Insufficient internet to download available dataset info. Check your internet connection.") from None
    return pd.read_csv(io.StringIO(dataset_link_text), header=0)

def download(url):
    try:
        dataset_text = _download_text(url)
    except NoInternetError:
        raise NoInternetError("Insufficient internet to download available dataset info. Check your internet connection.") from None
    return pd.read_csv(io.StringIO(dataset_text), header=0)

def get_patient_dataset_link():
    dataset_link_url = "https://bcm.box.com/shared/static/yt0ldc2nptxn96ovnna3hjxtlntzfvig.csv"
    try:
        dataset_link_text = _download_text(dataset_link_url)
    except NoInternetError:
        raise NoInternetError("Insufficient internet to download available dataset info. Check your internet connection.") from None
    return pd.read_csv(io.StringIO(dataset_link_text), header=0)


def download_patient_data(input):
    input = input.upper()
    data_link = get_patient_dataset_link()
    target = data_link.loc[data_link['Series'] == input]
    url = target.iloc[0,1]
    try:
        dataset_text = _download_text(url)
    except NoInternetError:
        raise NoInternetError("Insufficient internet to download available dataset info. Check your internet connection.") from None
    return pd.read_csv(io.StringIO(dataset_text), header=0)


""" ClinicalOmicsDB: Bridging the gap between next-generation clinical omics data and machine learning
Reference: Chang In Moon, Byron Jia, Bing Zhang
Paper link: TBD
Code author: Chang In Moon (moonchangin@gmail.com)
-----------------------------

preset_semi_supervised_models.py
- Train semi-supervised model and return predictions on the testing data

(1) stc_logit: self training logistic regression
(2) stc_xgb_model: self training XGBoost model
(3) stc_mlp: self training multi-layer perceptrons
(4) stc_svm: self training support vector machine classifier
(5) stc_knn: self training k-nearest neighbors classifier
(6) stc_rf: self training random forest classifier
(7) label_prop: Label propagation rbf kernel
(8) label_spread: Label spreading rbf kernel

"""

# Necessary packages
import pandas as pd 
import numpy as np
from sklearn.semi_supervised import LabelPropagation, LabelSpreading, SelfTrainingClassifier
from sklearn.linear_model import LogisticRegression
import xgboost as xgb
from sklearn.neural_network import MLPClassifier
from sklearn.svm import SVC
from sklearn.neighbors import KNeighborsClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.feature_selection import SelectPercentile, VarianceThreshold
from sklearn.model_selection import GridSearchCV
from sklearn.pipeline import Pipeline
from sklearn.decomposition import PCA
from sklearn.metrics.pairwise import check_pairwise_arrays, euclidean_distances

#%% 
def stc_logit(x_train, y_train, x_unlabel, x_test, parameters=None):
  """self training logistic regression.
  
  Args: 
    - x_train, y_train: training dataset
    - x_unlabel: unlabeled training dataset
    - x_test: testing feature
    - parameters: hyperparameter for optimization

  Returns:
    - y_test_hat: predicted values for x_test
  """
  # Define and fit model on training dataset
  if parameters != None:
      pipe = Pipeline([
      ('filter', VarianceThreshold()),
      ('kbest', SelectPercentile(percentile=20)),
      ('PCA', PCA(n_components = 10)),
      ('classifier', LogisticRegression(penalty='elasticnet', solver='saga', max_iter=10000, class_weight=True))
      ])
      model = GridSearchCV(pipe, parameters, scoring='roc_auc', cv=5, refit=True, n_jobs = 8, error_score=0)
      model.fit(x_train, y_train)
      self_training_model = SelfTrainingClassifier(model.best_estimator_, criterion = "threshold", max_iter = None, threshold = 0.6)
      X_train_mixed = pd.concat([x_train, x_unlabel])
      y_train_mixed = pd.concat([y_train, pd.Series([-1] * len(x_unlabel))])
      self_training_model.fit(X_train_mixed, y_train_mixed)
  elif parameters == None:    
      self_training_model = SelfTrainingClassifier(LogisticRegression(), criterion = "k_best", max_iter = None, threshold = 0.6)
      X_train_mixed = pd.concat([x_train, x_unlabel])
      y_train_mixed = pd.concat([y_train, pd.Series([-1] * len(x_unlabel))])
      self_training_model.fit(X_train_mixed, y_train_mixed)
  # Predict on x_test
  print(model.best_params_) 
  y_test_hat = self_training_model.predict_proba(x_test) 
  return y_test_hat

#%% 
def stc_xgb_model(x_train, y_train, x_unlabel, x_test, parameters=None):
  """self training XGBoost model
  
  Args: 
    - x_train, y_train: training dataset
    - x_unlabel: unlabeled training dataset
    - x_test: testing feature
    - parameters: hyperparameter for optimization

  Returns:
    - y_test_hat: predicted values for x_test
  """

  # Define and fit model on training dataset
  if parameters != None:
      pipe = Pipeline([
      ('filter', VarianceThreshold()),
      ('kbest', SelectPercentile(percentile=20)),
      ('PCA', PCA(n_components = 10)),
      ('classifier', xgb.XGBClassifier(verbosity = 0, silent=True, n_jobs = 8))
      ])
      model = GridSearchCV(pipe, parameters, scoring='roc_auc', cv=5, refit=True, n_jobs = 4, error_score=0)
      model.fit(x_train, y_train)
      self_training_model = SelfTrainingClassifier(model.best_estimator_, criterion = "threshold", max_iter = None, threshold = 0.6)
      X_train_mixed = pd.concat([x_train, x_unlabel])
      y_train_mixed = pd.concat([y_train, pd.Series([-1] * len(x_unlabel))])
      self_training_model.fit(X_train_mixed, y_train_mixed)
  elif parameters == None:
      self_training_model = SelfTrainingClassifier(xgb.XGBClassifier(verbosity = 0, silent=True, n_jobs = 8),
       criterion = "k_best", max_iter = None, threshold = 0.6)
      X_train_mixed = pd.concat([x_train, x_unlabel])
      y_train_mixed = pd.concat([y_train, pd.Series([-1] * len(x_unlabel))])
      self_training_model.fit(X_train_mixed, y_train_mixed)
  # Predict on x_test
  print(model.best_params_) 
  y_test_hat = self_training_model.predict_proba(x_test) 
  return y_test_hat


#%% 
def stc_mlp(x_train, y_train, x_unlabel, x_test, parameters=None):
  """self training multi-layer perceptrons
  
  Args: 
    - x_train, y_train: training dataset
    - x_unlabel: unlabeled training dataset
    - x_test: testing feature
    - parameters: hyperparameter for optimization

  Returns:
    - y_test_hat: predicted values for x_test
  """

  # Define and fit model on training dataset
  if parameters != None:
      pipe = Pipeline([
      ('filter', VarianceThreshold()),
      ('kbest', SelectPercentile(percentile=20)),
      ('PCA', PCA(n_components = 10)),
      ('classifier', MLPClassifier())
      ])
      model = GridSearchCV(pipe, parameters, scoring='roc_auc', cv=5, refit=True, n_jobs = 8, error_score=0)
      model.fit(x_train, y_train)
      self_training_model = SelfTrainingClassifier(model.best_estimator_, criterion = "threshold", max_iter = None, threshold = 0.6)
      X_train_mixed = pd.concat([x_train, x_unlabel])
      y_train_mixed = pd.concat([y_train, pd.Series([-1] * len(x_unlabel))])
      self_training_model.fit(X_train_mixed, y_train_mixed)
  elif parameters == None:
      self_training_model = SelfTrainingClassifier(MLPClassifier(),
       criterion = "k_best", max_iter = None, threshold = 0.6)
      X_train_mixed = pd.concat([x_train, x_unlabel])
      y_train_mixed = pd.concat([y_train, pd.Series([-1] * len(x_unlabel))])
      self_training_model.fit(X_train_mixed, y_train_mixed)
  # Predict on x_test
  print(model.best_params_) 
  y_test_hat = self_training_model.predict_proba(x_test) 
  return y_test_hat


#%% 
def stc_svm(x_train, y_train, x_unlabel, x_test, parameters=None):
  """self training support vector machine classifier
  
  Args: 
    - x_train, y_train: training dataset
    - x_unlabel: unlabeled training dataset
    - x_test: testing feature
    - parameters: hyperparameter for optimization

  Returns:
    - y_test_hat: predicted values for x_test
  """

  # Define and fit model on training dataset
  if parameters != None:
      pipe = Pipeline([
      ('filter', VarianceThreshold()),
      ('kbest', SelectPercentile(percentile=20)),
      ('PCA', PCA(n_components = 10)),
      ('classifier', SVC(probability = True))
      ])
      model = GridSearchCV(pipe, parameters, scoring='roc_auc', cv=5, refit=True, n_jobs = 8, error_score=0)
      model.fit(x_train, y_train)
      self_training_model = SelfTrainingClassifier(model.best_estimator_, criterion = "threshold", max_iter = None, threshold = 0.6)
      X_train_mixed = pd.concat([x_train, x_unlabel])
      y_train_mixed = pd.concat([y_train, pd.Series([-1] * len(x_unlabel))])
      self_training_model.fit(X_train_mixed, y_train_mixed)
  elif parameters == None:
      self_training_model = SelfTrainingClassifier(SVC(probability = True),
       criterion = "k_best", max_iter = None, threshold = 0.6)
      X_train_mixed = pd.concat([x_train, x_unlabel])
      y_train_mixed = pd.concat([y_train, pd.Series([-1] * len(x_unlabel))])
      self_training_model.fit(X_train_mixed, y_train_mixed)
  # Predict on x_test
  print(model.best_params_) 
  y_test_hat = self_training_model.predict_proba(x_test) 
  return y_test_hat
#%% 
def stc_knn(x_train, y_train, x_unlabel, x_test, parameters=None):
  """self training k-nearest neighbors classifier
  
  Args: 
    - x_train, y_train: training dataset
    - x_unlabel: unlabeled training dataset
    - x_test: testing feature
    - parameters: hyperparameter for optimization

  Returns:
    - y_test_hat: predicted values for x_test
  """

  # Define and fit model on training dataset
  if parameters != None:
      pipe = Pipeline([
      ('filter', VarianceThreshold()),
      ('kbest', SelectPercentile(percentile=20)),
      ('PCA', PCA(n_components = 10)),
      ('classifier', KNeighborsClassifier())
      ])
      model = GridSearchCV(pipe, parameters, scoring='roc_auc', cv=5, refit=True, n_jobs = 8, error_score=0)
      model.fit(x_train, y_train)
      self_training_model = SelfTrainingClassifier(model.best_estimator_, criterion = "threshold", max_iter = None, threshold = 0.6)
      X_train_mixed = pd.concat([x_train, x_unlabel])
      y_train_mixed = pd.concat([y_train, pd.Series([-1] * len(x_unlabel))])
      self_training_model.fit(X_train_mixed, y_train_mixed)
  elif parameters == None:
      self_training_model = SelfTrainingClassifier(KNeighborsClassifier(),
       criterion = "k_best", max_iter = None, threshold = 0.6)
      X_train_mixed = pd.concat([x_train, x_unlabel])
      y_train_mixed = pd.concat([y_train, pd.Series([-1] * len(x_unlabel))])
      self_training_model.fit(X_train_mixed, y_train_mixed)
  # Predict on x_test
  print(model.best_params_) 
  y_test_hat = self_training_model.predict_proba(x_test) 
  return y_test_hat

#%% 
def stc_rf(x_train, y_train, x_unlabel, x_test, parameters=None):
  """self training random forest
  
  Args: 
    - x_train, y_train: training dataset
    - x_unlabel: unlabeled training dataset
    - x_test: testing feature
    - parameters: hyperparameter for optimization

  Returns:
    - y_test_hat: predicted values for x_test
  """

  # Define and fit model on training dataset
  if parameters != None:
      pipe = Pipeline([
      ('filter', VarianceThreshold()),
      ('kbest', SelectPercentile(percentile=20)),
      ('PCA', PCA(n_components = 10)),
      ('classifier', RandomForestClassifier(n_jobs = 4))
      ])
      model = GridSearchCV(pipe, parameters, scoring='roc_auc', cv=5, refit=True, n_jobs = 8, error_score=0)
      model.fit(x_train, y_train)
      self_training_model = SelfTrainingClassifier(model.best_estimator_, criterion = "threshold", max_iter = None, threshold = 0.6)
      X_train_mixed = pd.concat([x_train, x_unlabel])
      y_train_mixed = pd.concat([y_train, pd.Series([-1] * len(x_unlabel))])
      self_training_model.fit(X_train_mixed, y_train_mixed)
  elif parameters == None:
      self_training_model = SelfTrainingClassifier(RandomForestClassifier(),
       criterion = "k_best", max_iter = None, threshold = 0.6)
      X_train_mixed = pd.concat([x_train, x_unlabel])
      y_train_mixed = pd.concat([y_train, pd.Series([-1] * len(x_unlabel))])
      self_training_model.fit(X_train_mixed, y_train_mixed)
  # Predict on x_test
  y_test_hat = self_training_model.predict_proba(x_test) 
  print(model.best_params_) 
  return y_test_hat  
#%% 
def label_prop(x_train, y_train, x_unlabel, x_test, parameters=None):
  """Label propagation RBF kernel
  
  Args: 
    - x_train, y_train: training dataset
    - x_unlabel: unlabeled training dataset
    - x_test: testing feature

  Returns:
    - y_test_hat: predicted values for x_test
  """
  pipe = Pipeline([
      ('filter', VarianceThreshold()),
      ('kbest', SelectPercentile(percentile=20)),
      ('PCA', PCA(n_components = 10)),
      ('classifier', LabelPropagation(kernel="rbf",
      max_iter=1000,
      gamma = 1e-3,
      tol=1e-3))
      ])
  model = GridSearchCV(pipe, parameters, scoring='roc_auc', cv=5, refit=True, n_jobs = 8, error_score=0)
  X_train_mixed = pd.concat([x_train, x_unlabel])
  y_train_mixed = pd.concat([y_train, pd.Series([-1] * len(x_unlabel))])
  model.fit(X_train_mixed, y_train_mixed)
  print(model.best_params_) 
  y_test_hat = model.predict_proba(x_test)
  return y_test_hat

#%% 
def label_spread(x_train, y_train, x_unlabel, x_test, parameters=None):
  """Label spreading RBF kernel
  
  Args: 
    - x_train, y_train: training dataset
    - x_unlabel: unlabeled training dataset
    - x_test: testing feature

  Returns:
    - y_test_hat: predicted values for x_test
  """
  pipe = Pipeline([
      ('filter', VarianceThreshold()),
      ('kbest', SelectPercentile(percentile=20)),
      ('PCA', PCA(n_components = 10)),
      ('classifier', LabelSpreading(kernel="rbf",
      alpha = 0.2,
      max_iter=1000,
      gamma = 1e-3,
      tol=1e-3))
      ])
  model = GridSearchCV(pipe, parameters, scoring='roc_auc', cv=5, refit=True, n_jobs = 8, error_score=0)
  X_train_mixed = pd.concat([x_train, x_unlabel])
  y_train_mixed = pd.concat([y_train, pd.Series([-1] * len(x_unlabel))])
  model.fit(X_train_mixed, y_train_mixed)
  print(model.best_params_) 
  y_test_hat = model.predict_proba(x_test)
  return y_test_hat